module MeterpreterDetection;

export {
	redef enum Notice::Type += {
		Meterpreter_Detected,
		Meterpreter_Header_Detected,
	};

	type conn_triplet: record {
		orig_h: addr;
		resp_h: addr;
		resp_p: port;
	};

	# Corelight option: bro.pkgs.corelight.meterpreter.config.notice_suppression_interval
	option notice_suppression_interval = 3600 sec;

	## Avoid immediate notices in case something is happening that is causing
	## lots of false positives.
	option popular_threshold = 25;

	# Header log fields.
	type HeaderInfo: record {
		start_time: time &log;
		id: conn_id &log;
		uid: string &log;
		protocol: string &log;
		guid: string &log;
		staged: bool &log;
		encrypted: bool &log;
	};

	# General log fields.
	type GeneralInfo: record {
		start_time: time &log;
		id: conn_id &log;
		uid: string &log;
		protocol: string &log;
		reason: string &log;
		os: string &log;
	};

	# Timing log fields.
	type TimingInfo: record {
		uid: string &log;
		sot: vector of double &log;
	};

	# For popularity filter
	type host: record {
		h: addr;
		p: port;
	};

	# To avoid OOM if too many meterpreter sessions are detected.
	global watchlist_max_size = 10000;

	global id_to_triplet: function(id: conn_id): conn_triplet;
	global log_a_notice: function(c: connection, message: string,
	    notice_type: Notice::Type, is_orig: bool);
	global max: function(a: count, b: count): count;

	type NoticeInfoVec: vector of Notice::Info;
	global watchlist_expire: function(t: table[host] of NoticeInfoVec, h: host)
	    : interval;
	global watchlist: table[host] of NoticeInfoVec &write_expire=1 hr
	    &expire_func=watchlist_expire;
	global allowlist: set[host] &write_expire=1 day;
	global connection_graph: table[host] of set[addr] &write_expire=1 hr;
}

# Convert connection ID to connection triplet.
function id_to_triplet(id: conn_id): conn_triplet
	{
	return conn_triplet($orig_h=id$orig_h, $resp_h=id$resp_h, $resp_p=id$resp_p);
	}

# Return the max of two counts
function max(a: count, b: count): count
	{
	return ( a > b ) ? a : b;
	}

function log_a_notice(c: connection, message: string, notice_type: Notice::Type,
    is_orig: bool)
	{
	# Check if this server is allow-listed.
	local src = host($h=c$id$orig_h, $p=c$id$orig_p);
	local dst = host($h=c$id$resp_h, $p=c$id$resp_p);
	local entry: host = is_orig ? dst : src;
	if ( entry in allowlist &&
	     notice_type != Meterpreter_Header_Detected )
		return;

	local notice_content = Notice::Info($note=notice_type, $conn=c,
	    $identifier=cat(c$id$orig_h, message),
	    $suppress_for=notice_suppression_interval, $msg=message);

	if ( popular_threshold == 0 ||
	     notice_type == Meterpreter_Header_Detected ||
	     (|watchlist| > watchlist_max_size && entry !in watchlist) )
		{
		NOTICE(notice_content);
		return;
		}

	if ( entry !in watchlist )
		watchlist[entry] = vector(notice_content);
	else
		watchlist[entry] += notice_content;
	}

event zeek_done()
	{
	for ( h, info in watchlist )
		watchlist_expire(watchlist, h);
	}

function watchlist_expire(t: table[host] of NoticeInfoVec, h: host): interval
	{
	# Detection confirmed, write to general log.
	for ( notice_content_index in t[h] )
		NOTICE(t[h][notice_content_index]);

	t[h] = vector();
	delete connection_graph[h];

	return 0 sec;
	}

function connection_established_handler(src: addr, dst: host)
	{
	if ( dst !in watchlist )
		return;

	if ( dst !in connection_graph )
		connection_graph[dst] = set();

	add connection_graph[dst][src];

	if ( |connection_graph[dst]| > popular_threshold )
		{
		delete watchlist[dst];
		delete connection_graph[dst];

		add allowlist[dst];
		}
	}

event connection_established(c: connection)
	{
	connection_established_handler(c$id$orig_h, host($h=c$id$resp_h, $p=c$id$resp_p));
	}
