module MeterpreterDetection_HTTPHeader;

export {
	# State of HTTP parsing.
	type parsing_state: enum { IDLE, FOUND_LENGTH };

	type HTTP_Info: record {
		reply_count: count &default=0;
		status: parsing_state &default=IDLE;
		length: count &default=0;
	};

	# These suspected and confirmed Meterpreter HTTP connections.
	global http_suspected: table[MeterpreterDetection::conn_triplet] of count
	    &create_expire=12 hr;
	global http_verified: set[MeterpreterDetection::conn_triplet]
	    &create_expire=12 hr;

	# State tracking for all HTTP connections.
	global http_info: table[conn_id] of HTTP_Info &create_expire=60 sec;

	# Type for HTTP check return.
	type Header: record {
		guid: string;
		encrypted: bool;
		len: count;
		request: bool;
	};

	# Authorized values in Meterpreter header.
	global allowed_encrypt_flags = set(0, 1);
	global allowed_mtpt_types = set(0, 1, 10, 11);
}

const http_body_analysis_byte_depth = 64 &redef;

redef record HTTP::Info += {
	meterpreter_analyzed_traffic_length: count &default=0;
};

# Functions for HTTP header filtering
function xor_string(a: string, b: string): count
	{
	return bytestring_to_count(a) ^ bytestring_to_count(b);
	}

# Parameter for reservoir sampling.
global sampling_size = 3;

function reservoir_sampling(cnt: count): bool
	{
	if ( cnt <= sampling_size )
		return T;

	return rand(cnt - 1) < sampling_size;
	}

# Takes a length and raw payload, returns the Meterpreter header information.
# If this is not a Meterpreter header, return the empty header.

const empty_header = Header($guid="", $encrypted=F, $len=0, $request=T);

function check_header(len: count, payload: string): Header
	{
	if ( len < 32 )
		return empty_header;

	local xor_key = payload[: 4];
	local encrypt_flag = xor_string(xor_key, payload[20 : 24]);
	local mtpt_type = xor_string(xor_key, payload[28 : 32]);

	if ( encrypt_flag in allowed_encrypt_flags &&
	     mtpt_type in allowed_mtpt_types )
		{
		local dbl_xor_key = xor_key + xor_key;
		local guid1 = MeterpreterDetection::count_to_bytestring(xor_string(dbl_xor_key, payload[4 : 12]));
		local guid2 = MeterpreterDetection::count_to_bytestring(xor_string(dbl_xor_key, payload[12 : 20]));
		local guid = guid1 + guid2;

		return Header($guid=guid,
			$encrypted=(encrypt_flag == 1),
			$len=xor_string(xor_key, payload[24 : 28]),
			$request=(mtpt_type == 0 || mtpt_type == 10));
		}

	return empty_header;
	}

function http_header_helper(cid: conn_id, value: string)
	{
	if ( MeterpreterDetection::id_to_triplet(cid) in http_verified )
		return;

	# Get length value.
	local content_length = 0;
	local stripped_value = strip(value);

	if ( |stripped_value| > 0 && is_num(stripped_value) )
		content_length = to_count(stripped_value);

	if ( content_length == 0 )
		return;

	if ( cid !in http_info )
		http_info[cid] = HTTP_Info();

	local rec = http_info[cid];

	if ( rec$status != IDLE )
		return;

	# Sampling - avoid checking every single packet in long connections.
	# If we decide to check this packet, switch to FOUND_LENGTH state.
	if ( reservoir_sampling(++rec$reply_count) )
		{
		rec$length = content_length;
		rec$status = FOUND_LENGTH;
		}
	}

# Catch Content-Length headers to check against the value in the MTPT header.
event http_header(c: connection, is_orig: bool, name: string, value: string)
	{
	if ( ! is_orig && name == "CONTENT-LENGTH" )
		http_header_helper(c$id, value);
	}

event http_entity_data_helper(f: fa_file, data: string)
	{
	for ( cid in f$conns )
		{
		local c: connection = f$conns[cid];

		if ( ! c?$http )
			next;

		if ( c$id !in http_info )
			next;

		local http = c$http;

		if ( http$meterpreter_analyzed_traffic_length >
		     http_body_analysis_byte_depth )
			return;

		local trip = MeterpreterDetection::id_to_triplet(c$id);
		if ( trip in http_verified )
			next;

		local rec = http_info[c$id];

		if ( rec$status != FOUND_LENGTH )
			next;

		rec$status = IDLE;

		# Check Meterpreter header.
		local length = |data|;
		local mtpt_hdr = check_header(length, data);
		local message: string;

		if ( mtpt_hdr$len + 24 == rec$length && mtpt_hdr$len > 0 )
			{
			# Raise Notice and add to suspected
			if ( trip !in http_suspected )
				{
				http_suspected[trip] = 0;
				message = fmt("Meterpreter detected with reason=HEADER_DETECTED, protocol=HTTP, os=ANY");
				MeterpreterDetection::log_a_notice(c, message,
				    MeterpreterDetection::Meterpreter_Detected, T);
				}

			# Check guid and encrypted flag, in case in an
			# initialization phase. Same logic as TCP case.
			local null_str = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
			if ( http_suspected[trip] > 10 ||
			     (mtpt_hdr$guid != null_str &&
			      (mtpt_hdr$encrypted || http_suspected[trip] > 3)) )
				{
				add http_verified[trip];
				message = fmt("Meterpreter header detected with protocol=HTTP in connection UID=%s, guid=%s, staged=T,"
				    + "encrypted=%s", c$uid, string_to_ascii_hex(mtpt_hdr$guid), mtpt_hdr$encrypted);
				MeterpreterDetection::log_a_notice(c, message, MeterpreterDetection::Meterpreter_Header_Detected,
				    T);
				}

			else
				{
				++http_suspected[trip];
				--rec$reply_count;
				rec$status = IDLE;
				}
			}

		http$meterpreter_analyzed_traffic_length += |data|;
		if ( http$meterpreter_analyzed_traffic_length >
		     http_body_analysis_byte_depth )
			Files::remove_analyzer(f, Files::ANALYZER_DATA_EVENT,
				[$stream_event=http_entity_data_helper]);
		}
	}

event file_over_new_connection(f: fa_file, c: connection, is_orig: bool)
	{
	if ( ! is_orig && c?$http && c$http?$method )
		Files::add_analyzer(f, Files::ANALYZER_DATA_EVENT,
			[$stream_event=http_entity_data_helper]);
	}
