module MeterpreterDetection_Timing;

export {
	global verified: set[conn_id] &create_expire=12 hr;
	global threshold = 15;

	# Three states of the state machine.
	type states: enum { NOISE, LINEAR, CONSTANT };

	# Polling history of each connection.
	type polling_state: record {
		current_index: count &default=0;
		max_index: count &default=0;

		last_delta: interval &default=0 sec;
		last_timestamp: time &default=current_time();
		last_size: count &default=0;

		state: states &default=NOISE;
		last_delay: interval &default=0 sec;

		slope_sum: interval &default=0 sec;
	};

	# Parameters for capture. These have been fine-tuned to avoid FPs.
	type constant_values: record {
		base: interval &default=0.1 sec;
		end: interval &default=10 sec;

		instant_tolerance: interval &default=25 msec;
		average_tolerance: interval &default=2 msec;

		max_capture_start: interval &default=2 sec;
	};

	global win_polling_data:
		table[conn_id] of polling_state &write_expire=60 sec;
	global flagged:
		set[MeterpreterDetection::conn_triplet] &create_expire=12 hr;
	global constants: constant_values;

	global student_quantiles =
		vector(700, 32, 13, 9, 7, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 4);
}

function abs_interval(a: interval): interval
	{
	return a > 0 sec ? a : -a;
	}

# Check if current sequence of slopes for inter-arrivals is close to 100 ms.
function check_value(slope: interval, slope_sum: interval, n: count): bool
	{
	# Check if value is in acceptable bounds.
	local diff = slope - constants$base;

	# Instant check. Is the current slope not close enough?
	if ( abs_interval(diff) > constants$instant_tolerance )
		return F;

	# Average check. Is the average value of the slopes close to 100ms ?
	if ( n < 2 )
		return T;

	local k = n < 18 ? student_quantiles[n - 2] : 4;
	local confidence_interval = k * 1 msec / sqrt(n - 1);
	local average_error = (slope_sum + diff) / n;

	return average_error <= constants$average_tolerance + confidence_interval;
	}

# State machine logic.
function next_step(current_state: polling_state, ts: time, poll_size: count,
    uid: string): polling_state
	{
	local delta = ts - current_state$last_timestamp - current_state$last_delay;
	local slope = delta - current_state$last_delta;
	local check = check_value(slope, current_state$slope_sum,
					current_state$current_index + 1);

	if ( current_state$state == NOISE )
		{
		if ( check && poll_size == current_state$last_size &&
		     delta < constants$max_capture_start &&
		     delta > constants$base - constants$instant_tolerance )
			{ # NOISE -> LINEAR
			current_state$current_index = 1;
			current_state$state = LINEAR;
			current_state$slope_sum += slope - constants$base;
			}
		}

	else if ( current_state$state == LINEAR )
		{
		if ( check && poll_size == current_state$last_size &&
		     delta < constants$end + constants$instant_tolerance &&
		     delta > constants$base )
			{ # LINEAR -> LINEAR
			++current_state$current_index;
			current_state$slope_sum += slope - constants$base;
			}

		else if ( poll_size == current_state$last_size &&
			  abs_interval(delta - constants$end) < constants$instant_tolerance )
			{ # LINEAR -> CONSTANT
			++current_state$current_index;
			current_state$slope_sum = 0 sec;
			current_state$state = CONSTANT;
			}

		else
			{ # LINEAR -> NOISE
			current_state$max_index = MeterpreterDetection::max(current_state$max_index,
			    current_state$current_index);
			current_state$current_index = 0;
			current_state$state = NOISE;
			current_state$slope_sum = 0 sec;
			}
		}

	else
		{
		if ( poll_size == current_state$last_size &&
		    abs_interval(delta - constants$end) < constants$instant_tolerance )
			# CONSTANT -> CONSTANT
			++current_state$current_index;

		else
			{ # CONSTANT -> NOISE
			current_state$max_index = MeterpreterDetection::max(current_state$max_index,
			    current_state$current_index);
			current_state$current_index = 0;
			current_state$state = NOISE;
			current_state$slope_sum = 0 sec;
			}
		}

	current_state$last_delta = delta;
	current_state$last_timestamp = ts;
	current_state$last_size = poll_size;

	return current_state;
	}
